@@ -1,4 +1,4 @@
-from lhub import common, exceptions, log
+from lhub import common, exceptions
 from lhub.lhub import LogicHub
 from lhub.api import LogicHubAPI
 from lhub.url import URLs
