@@ -3,7 +3,7 @@ import json
 import re
 from collections import namedtuple
 
-from requests import request
+from requests import request, models
 from requests.exceptions import HTTPError
 from urllib3 import disable_warnings
 from urllib3.exceptions import InsecureRequestWarning
@@ -23,6 +23,7 @@ cached_obj = namedtuple('CachedObject', ['time', 'value'])
 
 class LogicHubAPI:
     DEFAULT_ALERT_RESULT_LIMIT = 100
+    DEFAULT_CASE_RESULT_LIMIT = 25
     DEFAULT_CACHE_SECONDS = 300
     USER_AGENT_STRING = "LHUB CLI"
     http_timeout_default = 120
@@ -229,11 +230,33 @@ class LogicHubAPI:
             return {}
         return {'PLAY_SESSION': self.session_cookie}
 
-    def __standard_http_response_tests(self, response_obj, url):
+    def __standard_http_response_tests(self, response_obj: models.Response, url: str, input_var=None):
+        # If the response is okay, return successfully right away
+        try:
+            response_obj.raise_for_status()
+            return
+        except HTTPError as e:
+            caught_exception = e
+
+        try:
+            __response_json = response_obj.json()
+        except json.decoder.JSONDecodeError:
+            __response_json = {}
+
+        __status_code = response_obj.status_code
+        __errors = __response_json.get("errors", [])
+        __primary_error_type = __errors[0].get("errorType") if __errors else None
+        exception_inputs = {
+            "input_var": input_var,
+            "url": url,
+            "last_response_status": __status_code,
+            "last_response_text": response_obj.text,
+        }
+
         # Group all tests for status code 401
-        if response_obj.status_code == 401:
+        if __status_code == 401:
             if self.auth_type == 'password':
-                raise exceptions.auth.PasswordAuthFailure
+                raise exceptions.auth.PasswordAuthFailure(**exception_inputs)
             else:
                 if 'Unauthorized for URL' in response_obj.text or 'token is not allowed with this endpoint' in response_obj.text:
                     raise exceptions.auth.APIAuthNotAuthorized(url)
@@ -241,7 +264,7 @@ class LogicHubAPI:
                     raise exceptions.auth.APIAuthFailure(url)
 
         # Group all tests for status code 400
-        elif response_obj.status_code == 400:
+        elif __status_code == 400:
             if self.auth_type == 'password':
                 if url == self.url.login and 'Username/Password not valid' in self.last_response_text:
                     raise exceptions.auth.PasswordAuthFailure
@@ -249,20 +272,25 @@ class LogicHubAPI:
                 _id = re.search(r'POST /demo/batch-(\d+)', response_obj.text)
                 _id = _id.groups()[0] if _id else None
                 raise exceptions.app.BatchNotFound(_id)
+            if __primary_error_type == "AlreadyPresentException":
+                if url == self.url.user_create:
+                    raise exceptions.app.UserAlreadyExists(user=input_var)
+                if url == self.url.user_group_create:
+                    raise exceptions.app.UserGroupAlreadyExists(input_var=input_var)
 
         # Group all tests for status code 500
-        elif response_obj.status_code == 500:
+        elif __status_code == 500:
             if 'Unable to find batch with id' in response_obj.text:
                 _id = re.search(r'Unable to find batch with id batch-(\d+)', response_obj.text)
                 _id = _id.groups()[0] if _id else None
                 raise exceptions.app.BatchNotFound(_id)
 
-        response_obj.raise_for_status()
+        raise caught_exception
 
     def _http_request(
             self, url, method="GET", params: dict = None, body=None,
             headers: dict = None, timeout=None, test_response: bool = True,
-            reauth=True, **kwargs):
+            reauth=True, input_var=None, **kwargs):
         method = method.upper() if method else "GET"
         # Only use reauth/automatic login if password auth is used
         if self.auth_type == 'password':
@@ -298,10 +326,11 @@ class LogicHubAPI:
                 response = request(method=method, url=url, verify=self.verify_ssl, timeout=timeout, cookies=self.cookies, **kwargs)
 
         # Store last response before testing whether the call was successful
+        self.last_url = url
         self.last_response_status = response.status_code
         self.last_response_text = response.text
         if test_response:
-            self.__standard_http_response_tests(response, url)
+            self.__standard_http_response_tests(response, url, input_var=input_var)
         return response
 
     def login(self, force_new_session=False):
@@ -315,7 +344,7 @@ class LogicHubAPI:
             return
 
         log.debug("Logging into LogicHub")
-        login_response = self._http_request(url=self.url.login, method="POST", body=self.__credentials, timeout=self._http_timeout_login, reauth=False)
+        login_response = self._http_request(url=self.url.login, method="POST", body=self.__credentials, timeout=self._http_timeout_login, reauth=False, input_var=self.__username)
         self.session_cookie = login_response.cookies.get("PLAY_SESSION")
         log.debug("Login successful")
         _ = atexit.register(self.close)
@@ -329,7 +358,7 @@ class LogicHubAPI:
             return
 
         log.debug("Issuing logout command")
-        _response = self._http_request(url=self.url.logout, method="POST", test_response=False, reauth=False, timeout=self._http_timeout_login)
+        _response = self._http_request(url=self.url.logout, method="POST", test_response=False, reauth=False, timeout=self._http_timeout_login, input_var=self.__username)
         try:
             _response.raise_for_status()
         except HTTPError as err:
@@ -382,7 +411,7 @@ class LogicHubAPI:
         offset = int(offset or 0)
         params = {"fields": "*", "pageSize": limit, "after": offset, "cachedOnly": "true"}
         log.debug("Fetching batch results")
-        response = self._http_request(url=self.url.batch_results_by_id.format(int(batch_id)), params=params)
+        response = self._http_request(url=self.url.batch_results_by_id.format(int(batch_id)), params=params, input_var=batch_id)
         return response.json()
 
     def get_batches_by_stream_id(self, stream_id, limit=None, offset=0, statuses=None, exclude_empty_results=False):
@@ -402,7 +431,7 @@ class LogicHubAPI:
         }
         body = {"status": statuses or [], "excludeBatchesWithZeroEvents": exclude_empty_results}
         log.debug("Fetching batches")
-        response = self._http_request(method="POST", url=self.url.stream_batches.format(stream_id), params=params, body=body)
+        response = self._http_request(method="POST", url=self.url.stream_batches.format(stream_id), params=params, body=body, input_var=stream_id)
         return response.json()
 
     def get_connection_status(self, connections: list):
@@ -458,6 +487,7 @@ class LogicHubAPI:
             params={"offset": offset, "limit": limit},
             body=payload,
             headers={'Content-Type': 'application/json'},
+            input_var=list_id
         )
         results = response.json()
         try:
@@ -469,19 +499,19 @@ class LogicHubAPI:
     # ToDo Token auth not supported as of 2022-05-05 (m94)
     def get_dashboard(self, dashboard_id):
         log.debug("Fetching dashboard data")
-        response = self._http_request(url=self.url.dashboard.format(dashboard_id))
+        response = self._http_request(url=self.url.dashboard.format(dashboard_id), input_var=dashboard_id)
         return response.json()
 
     # ToDo Token auth not supported as of 2022-05-05 (m94)
     def get_dashboard_data(self, dashboard_id):
         log.debug("Fetching dashboard data")
-        response = self._http_request(url=self.url.dashboard_data.format(dashboard_id))
+        response = self._http_request(url=self.url.dashboard_data.format(dashboard_id), input_var=dashboard_id)
         return response.json()
 
     # ToDo Token auth not supported as of 2022-05-05 (m94)
     def get_widget_data(self, dashboard_id, widget_id):
         log.debug("Fetching dashboard data")
-        response = self._http_request(url=self.url.widget_data.format(dashboard_id=dashboard_id, widget_id=widget_id))
+        response = self._http_request(url=self.url.widget_data.format(dashboard_id=dashboard_id, widget_id=widget_id), input_var=widget_id)
         return response.json()
 
     # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
@@ -496,12 +526,11 @@ class LogicHubAPI:
 
     # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
     def get_rules_for_rule_set(self, rule_set):
-        rule_set = helpers.format_rule_set_id(rule_set)
         params = {
             "fields": "name,isPublic,rules[filter,values*1000,score]*1000",
         }
         log.debug("Fetching rules")
-        response = self._http_request(url=self.url.rule_set.format(rule_set), params=params)
+        response = self._http_request(url=self.url.rule_set.format(helpers.format_rule_set_id(rule_set)), params=params, input_var=rule_set)
         results = response.json()
         try:
             results = results["result"]["rules"]["data"]
@@ -512,7 +541,7 @@ class LogicHubAPI:
     def get_stream_by_id(self, stream_id: int):
         headers = {"Accept": "application/json"}
         log.debug("Fetching stream")
-        response = self._http_request(url=self.url.stream_by_id.format(stream_id), headers=headers, test_response=False)
+        response = self._http_request(url=self.url.stream_by_id.format(stream_id), headers=headers, test_response=False, input_var=stream_id)
         if response.status_code == 400 and 'Cannot find entity for id StreamId' in response.text:
             raise exceptions.app.StreamNotFound(stream_id)
         else:
@@ -538,7 +567,7 @@ class LogicHubAPI:
     def get_workflow_by_id(self, workflow_id: int):
         assert isinstance(workflow_id, int), "Workflow ID must be an integer"
         log.debug("Fetching workflow")
-        response = self._http_request(method="GET", url=self.url.case_status_workflow_by_id.format(workflow_id))
+        response = self._http_request(method="GET", url=self.url.case_status_workflow_by_id.format(workflow_id), input_var=workflow_id)
         return response.json()
 
     # ToDo STILL DOES NOT WORK WITH API AUTH AS OF M91
@@ -556,7 +585,7 @@ class LogicHubAPI:
             }
         }
         log.debug("Issuing new scoring rule")
-        response = self._http_request(url=self.url.rule_set.format(rule_set), method="POST", body=body, headers=headers)
+        response = self._http_request(url=self.url.rule_set.format(rule_set), method="POST", body=body, headers=headers, input_var=rule_set)
         results = response.json()
         try:
             results = results["result"]
@@ -566,7 +595,7 @@ class LogicHubAPI:
 
     def export_playbook(self, playbook_id, test_response=True):
         log.debug(f"Exporting playbook {playbook_id}")
-        response = self._http_request(url=self.url.flow_export.format(playbook_id), test_response=test_response)
+        response = self._http_request(url=self.url.flow_export.format(playbook_id), test_response=test_response, input_var=playbook_id)
         return response.json()
 
     def execute_command(self, command_payload, limit=None):
@@ -588,8 +617,9 @@ class LogicHubAPI:
         try:
             result_dict = response.json()
         except json.decoder.JSONDecodeError:
-            log.fatal(f"Failed to load API response as JSON. Status code: {response.status_code} Response text: {response.text}")
-            sys.exit(1)
+            msg = f"Failed to load API response as JSON. Status code: {response.status_code} Response text: {response.text}"
+            log.fatal(msg)
+            raise exceptions.app.UnexpectedOutput(msg)
 
         errors = result_dict.pop("errors", [])
         if errors:
@@ -882,13 +912,18 @@ class LogicHubAPI:
             raise exceptions.app.UnexpectedOutput("API response does not match the expected schema for listing rule sets")
         return results
 
+    def list_saml_configs(self):
+        log.debug("Fetching SSO configurations")
+        response = self._http_request(url=self.url.saml_configs)
+        return response.json()
+
     def list_stream_states(self, stream_ids: List[int]):
         new_stream_id_list = []
         for s in stream_ids:
             new_stream_id_list.append(f'stream-{s}')
         body = {"streams": new_stream_id_list}
         log.debug("Fetching stream states")
-        response = self._http_request(method="POST", url=self.url.stream_states, body=body)
+        response = self._http_request(method="POST", url=self.url.stream_states, body=body, input_var=stream_ids)
         return response.json()
 
     def list_streams(self, search_text: str = None, filters: list = None, limit: int = None, offset: int = 0):
@@ -937,12 +972,13 @@ class LogicHubAPI:
         )
         return response.json()
 
-    def list_users(self, limit=None, hide_inactive=True):
+    def list_users_v1_up_to_m95(self, limit=None, hide_inactive=True):
         limit = limit if limit and isinstance(limit, int) else 99999
         params = {"pageSize": limit, "after": 0}
         body = {"filters": []}
         if hide_inactive:
             body['filters'].append({"hideInactive": True})
+        log.debug(f"TEMP -- body: {json.dumps(body)}")
         log.debug("Fetching users")
         response = self._http_request(
             url=self.url.users,
@@ -953,6 +989,33 @@ class LogicHubAPI:
         )
         return response.json()
 
+    def list_users_v2_from_m96(self, limit=None, **kwargs):
+        limit = limit if limit and isinstance(limit, int) else 99999
+        params = {"pageSize": limit, "after": 0}
+        body = {"filters": []}
+        if kwargs:
+            body["filters"].append(kwargs)
+        log.debug(f"TEMP -- body: {json.dumps(body)}")
+        log.debug("Fetching users")
+        response = self._http_request(
+            url=self.url.users,
+            method="POST",
+            headers={"Content-Type": "application/json"},
+            params=params,
+            body=body
+        )
+        return response.json()
+
+    def list_users(self, limit=None, **kwargs):
+        if self.major_version > 95 and "hide_inactive" in kwargs:
+            raise DeprecationWarning("The \"hideInactive\" filter is no longer valid since LogicHub version m96")
+
+        if self.major_version <= 95:
+            func = self.list_users_v1_up_to_m95
+        else:
+            func = self.list_users_v2_from_m96
+        return func(limit=limit, **kwargs)
+
     def list_workflows(self):
         log.debug("Fetching workflows")
         response = self._http_request(method="GET", url=self.url.case_status_list_workflows)
@@ -970,9 +1033,10 @@ class LogicHubAPI:
         headers = {"Content-Type": "application/json"}
         body = {"method": "reschedule", "parameters": {}}
         log.debug("Issuing command to reprocess batch")
-        response = self._http_request(method="POST", url=self.url.batch_reprocess.format(batch_id), headers=headers, body=body)
+        response = self._http_request(method="POST", url=self.url.batch_reprocess.format(batch_id), headers=headers, body=body, input_var=batch_id)
         return response.json()
 
+    # introduced somewhere between m70 and m80
     def me(self):
         log.debug("Fetching current user info")
         response = self._http_request(method="GET", url=self.url.me, timeout=30)
@@ -1001,7 +1065,7 @@ class LogicHubAPI:
 
     def update_current_user_preferences(self, preferences):
         log.debug("Patching user (updating preferences)")
-        response = self._http_request(method="PATCH", url=self.url.user.format(self.session_user_id), body={"preferences": preferences})
+        response = self._http_request(method="PATCH", url=self.url.user.format(self.session_user_id), body={"preferences": preferences}, input_var=self.session_user_id)
         return response.json()
 
     def case_prefix_refresh(self):
@@ -1029,10 +1093,23 @@ class LogicHubAPI:
     #     response = self._http_request(
     #         method="PATCH",
     #         url=self.url.case_update_linked_alerts.format(case_id=case_id),
-    #         headers={"Content-Type": "application/json"}
+    #         headers={"Content-Type": "application/json"},
+    #         input_var=case_id,
     #     )
     #     output = response.json()
 
+    def case_update(self, case_id: str, **kwargs):
+        if not kwargs:
+            raise exceptions.validation.InputValidationError("No changes provided for updating case", input_var=case_id, action_description="case update")
+        log.debug("Updating case")
+        response = self._http_request(
+            method="PATCH",
+            url=self.url.case.format(case_id=case_id),
+            headers={"Content-Type": "application/json"},
+            body=kwargs
+        )
+        return response.json()
+
     def cases_get_prefix(self):
         log.debug("Fetching case prefix")
         response = self._http_request(
@@ -1053,7 +1130,8 @@ class LogicHubAPI:
             url=self.url.notebooks_attached,
             method="POST",
             headers={"Content-Type": "application/json"},
-            body=body
+            body=body,
+            input_var=case_id
         )
         response = response.json()
         # Sort the results by notebook ID
@@ -1079,6 +1157,54 @@ class LogicHubAPI:
             method="PATCH",
             headers={"Content-Type": "application/json"},
             params={"libraryView": "all"},
+            body=body,
+            input_var=case_id
+        )
+        return response.json()
+
+    # ToDo Update alert search and validation methods to match the updates I made here
+    def cases_search_validate(self, query: str):
+        if query:
+            # First confirm that the query is valid
+            body = {"queryType": "advanced", "query": query}
+            log.debug("Validating search syntax")
+            response = self._http_request(
+                url=self.url.cases_search_validate,
+                method="POST",
+                headers={"Content-Type": "application/json"},
+                body=body
+            )
+            response = response.json()
+            is_invalid = response['result'].get('error')
+            if is_invalid is None:
+                raise exceptions.app.UnexpectedOutput('Unexpected response while validating query string.')
+            elif is_invalid:
+                raise exceptions.app.CaseQueryValidationError(response['result'].get('message'))
+
+    # ToDo Update alert search and validation methods to match the updates I made here
+    def case_search_advanced(self, query: str = None, limit: int = None, **kwargs):
+        # Validate inputs
+        if limit is None:
+            limit = self.DEFAULT_CASE_RESULT_LIMIT
+        elif limit == -1:
+            limit = 999999
+        elif not isinstance(limit, int):
+            raise exceptions.validation.InputValidationError("Invalid input type for search limit", input_var=limit)
+
+        body = {
+            "pageNumber": 0,
+            "pageSize": limit,
+            "query": query or "",
+            "includeTaskEntityType": True,
+            "includeWorkflow": True
+        }
+        if kwargs:
+            body.update(kwargs)
+        log.debug("Executing case search")
+        response = self._http_request(
+            url=self.url.cases_search_advanced,
+            method="POST",
+            headers={"Content-Type": "application/json"},
             body=body
         )
         return response.json()
@@ -1088,12 +1214,12 @@ class LogicHubAPI:
         url = self.url.alert_fetch.format(alert_id)
         # Send request, but disable immediate testing in order to do specialized testing first
         log.debug("Fetching alert")
-        response = self._http_request(method="GET", url=url, test_response=False, **kwargs)
+        response = self._http_request(method="GET", url=url, test_response=False, **kwargs, input_var=alert_id)
         if return_raw:
             return response
         if response.status_code == 400 and re.search(r'Alert with id alert-\d+ doesn\'t exist', response.text):
             raise ValueError(f'No alert with ID {alert_id}')
-        self.__standard_http_response_tests(response, url)
+        self.__standard_http_response_tests(response, url, input_var=alert_id)
         return response.json()
 
     def alerts_search_validate(self, query: str):
@@ -1112,9 +1238,9 @@ class LogicHubAPI:
             if is_invalid is None:
                 raise exceptions.app.UnexpectedOutput('Unexpected response while validating query string.')
             elif is_invalid:
-                raise exceptions.validation.AlertQueryValidationError(response['result'].get('message'))
+                raise exceptions.app.AlertQueryValidationError(response['result'].get('message'))
 
-    def alerts_search_advanced(self, query: str = None, limit: int = None):
+    def alerts_search_advanced(self, query: str = None, limit: int = None, **kwargs):
         # Sanitize inputs
         limit = int(limit) if limit and int(limit) > 0 else self.DEFAULT_ALERT_RESULT_LIMIT
         query = query.strip() if query and query.strip() else ""
@@ -1128,6 +1254,8 @@ class LogicHubAPI:
             "sortOrder": "desc",
             "query": query
         }
+        if kwargs:
+            body.update(kwargs)
         log.debug("Searching alerts")
         response = self._http_request(
             url=self.url.alerts_search_advanced,
@@ -1141,14 +1269,14 @@ class LogicHubAPI:
         # Sanitize stream ID
         body = [f"stream-{helpers.format_stream_id(stream_id)}"]
         log.debug("Pausing stream")
-        response = self._http_request(method="POST", url=self.url.stream_pause, body=body, **kwargs)
+        response = self._http_request(method="POST", url=self.url.stream_pause, body=body, **kwargs, input_var=stream_id)
         return response.json()
 
     def stream_resume(self, stream_id, **kwargs):
         # Sanitize stream ID
         body = [f"stream-{helpers.format_stream_id(stream_id)}"]
         log.debug("Resuming stream")
-        response = self._http_request(method="POST", url=self.url.stream_resume, body=body, **kwargs)
+        response = self._http_request(method="POST", url=self.url.stream_resume, body=body, **kwargs, input_var=stream_id)
         return response.json()
 
     def create_user(self, username, email, authentication_type, group_ids: list):
@@ -1159,12 +1287,34 @@ class LogicHubAPI:
             "authenticationType": authentication_type or "password"
         }
         log.debug(f"Creating user: {username}")
-        response = self._http_request(method="POST", url=self.url.user_create, body=body)
+        response = self._http_request(method="POST", url=self.url.user_create, body=body, input_var=username)
         return response.json()
 
     def delete_user(self, user_ids: list):
         log.debug(f"Deleting users by id: {', '.join([str(i) for i in user_ids])}")
-        response = self._http_request(method="POST", url=self.url.user_delete, body=user_ids)
+        response = self._http_request(method="POST", url=self.url.user_delete, body=user_ids, input_var=user_ids)
+        return response.json()
+
+    def create_user_group(self, name: str, users: list = None, **kwargs):
+        body = {"name": name, "users": users or [], **kwargs}
+        log.debug(f"Creating user: {name}")
+        response = self._http_request(method="POST", url=self.url.user_group_create, body=body, input_var=name)
+        return response.json()
+
+    def update_user_group(self, group_id: int, name: str, users: list = None, **kwargs):
+        body = {"name": name, "users": users or [], **kwargs}
+        log.debug(f"Creating user: {name}")
+        response = self._http_request(method="PATCH", url=self.url.user_group.format(group_id), body=body, input_var=name)
+        return response.json()
+
+    def get_password_settings(self):
+        log.debug("Fetching password settings")
+        response = self._http_request(url=self.url.get_password_settings)
+        return response.json()
+
+    def update_password_settings(self, settings_dict):
+        log.debug("Fetching password settings")
+        response = self._http_request(method="PATCH", url=self.url.update_password_settings, body=settings_dict)
         return response.json()
 
 
@@ -1173,11 +1323,18 @@ class FormattedObjects:
         self.__api = api
 
     @property
-    def system_field_lh_linked_alerts(self):
+    def version(self):
+        return super(self.__api).version
+
+    @property
+    def system_field_lh_linked_alerts(self) -> dict:
+        field = None
         for f in self.__api.fields['result']['data']:
             if f.get('fieldName') == 'lh_linked_alerts':
-                return f
-        raise exceptions.validation.VersionMinimumNotMet(min_version='m86', feature_label='linked alerts')
+                field = f
+                break
+        return field
+        # raise exceptions.validation.VersionMinimumNotMet(min_version='m86', feature_label='linked alerts')
 
     @property
     def system_field_lh_linked_alerts_id(self):
@@ -1211,7 +1368,11 @@ class FormattedObjects:
 
     @property
     def users(self):
-        response = self.__api.list_users(limit=None, hide_inactive=True)
+        if self.__api.major_version < 96:
+            kwargs = {"hide_inactive": True}
+        else:
+            kwargs = {"inactiveUsers": "onlyActive"}
+        response = self.__api.list_users(limit=None, **kwargs)
         return response['result']['data']
 
     @property
